package at.tugraz.iaik.auth.impl

import at.tugraz.iaik.auth.mgmt.config.Validity
import at.tugraz.iaik.auth.socket.AuthSocket
import at.tugraz.iaik.auth.structs.AuthRequestListener
import at.tugraz.iaik.auth.structs.AuthResponse
import at.tugraz.iaik.auth.structs.Request
import at.tugraz.iaik.p2p.commons.CommonCrypto
import at.tugraz.iaik.p2p.commons.CryptoException
import com.google.zxing.BarcodeFormat
import com.google.zxing.EncodeHintType
import com.google.zxing.qrcode.QRCodeWriter
import com.google.zxing.qrcode.decoder.ErrorCorrectionLevel
import iaik.asn1.structures.AlgorithmID
import iaik.security.ssl.CipherSuite
import java.awt.BorderLayout
import java.awt.Color
import java.awt.Graphics2D
import java.awt.image.BufferedImage
import java.io.IOException
import java.io.ObjectInputStream
import java.net.Socket
import java.security.PrivateKey
import java.security.cert.X509Certificate
import java.util.*
import javax.swing.*


/**
 * Created by Bernd Prünster <bernd.pruenster@iaik.tugraz.at>
 */
class QRAuth @Throws(IOException::class, CryptoException::class) constructor(port: Int,
                                                                             certChain: Array<X509Certificate>,
                                                                             cipherSuites: Array<CipherSuite>,
                                                                             validity: Validity,
                                                                             crlDistPoints: Array<String>?,
                                                                             signingKey: PrivateKey,
                                                                             sigAlg: AlgorithmID,
                                                                             clientTls: Boolean = false) : AuthSocket(
        port,
        null,
        AuthResponder(signingKey, certChain[0], sigAlg, validity, crlDistPoints),
        certChain,
        cipherSuites,
        signingKey) {
    init {
        server.context.requestClientCertificate = clientTls
    }

    private class AuthResponder(private val signingKey: PrivateKey,
                                private val caCert: X509Certificate,
                                private val sigAlg: AlgorithmID,
                                private val validity: Validity,
                                private val crlDistPoints: Array<String>?) : AuthRequestListener {


        @Throws(CryptoException::class) override fun onRequest(src: Socket,
                                                               auth: AuthSocket,
                                                               request: Request): AuthResponse {


            val token = CommonCrypto.RND.nextLong().toString(26)
            val popup = AuthPopup(token)

            val provided = ObjectInputStream(src.getInputStream()).readObject() as String?
            if (provided != token) return AuthResponse(AuthResponse.Result.DENIED)
            popup.isVisible = false
            popup.dispose()

            val to = Date()
            to.time += validity.unit.toMillis(validity.number.toLong())
            val cert = CommonCrypto.genCert(request.pubKey,
                                            request.principal,
                                            caCert.subjectDN,
                                            signingKey,
                                            sigAlg,
                                            false,
                                            Date(),
                                            to,
                                            crlDistPoints)
            return AuthResponse(AuthResponse.Result.GRANTED, cert)
        }
    }

    private class AuthPopup(token: String) : JFrame() {
        init {
            val hints = Hashtable<EncodeHintType, ErrorCorrectionLevel>()
            hints.put(EncodeHintType.ERROR_CORRECTION, ErrorCorrectionLevel.H)
            val qrCodeWriter = QRCodeWriter()
            val byteMatrix = qrCodeWriter.encode(token, BarcodeFormat.QR_CODE, 400, 400, hints)
            // Make the BufferedImage that are to hold the QRCode
            val matrixWidth = byteMatrix.width
            val image = BufferedImage(matrixWidth, matrixWidth, BufferedImage.TYPE_INT_RGB)
            image.createGraphics()
            val graphics = image.getGraphics() as Graphics2D;
            graphics.color = Color.WHITE;
            graphics.fillRect(0, 0, matrixWidth, matrixWidth);
            // Paint and save the image using the ByteMatrix
            graphics.color = Color.BLACK;

            for (i in 0 until matrixWidth) {
                for (j in 0 until matrixWidth) {
                    if (byteMatrix.get(i, j)) {
                        graphics.fillRect(i, j, 1, 1);
                    }
                }
            }
            val img = JLabel(ImageIcon(image))
            img.background = Color.WHITE
            val layout = BorderLayout()
            this.layout = layout
            add(img, BorderLayout.CENTER)
            val bot = JTextField(token)
            bot.horizontalAlignment = JTextField.CENTER
            bot.border = null
            bot.isOpaque = true
            bot.background = Color.WHITE
            bot.isEditable = false;
            bot.foreground = Color.BLACK
            add(bot, BorderLayout.SOUTH)
            val top = JLabel("Send the following Token", SwingConstants.CENTER)
            top.background = Color.WHITE
            top.isOpaque = true
            add(top, BorderLayout.NORTH)
            defaultCloseOperation = WindowConstants.DO_NOTHING_ON_CLOSE
            pack()
            setLocationRelativeTo(null)
            isVisible = true
        }
    }
}