package at.tugraz.iaik.auth.impl

import at.tugraz.iaik.auth.mgmt.config.Validity
import at.tugraz.iaik.auth.socket.AuthSocket
import at.tugraz.iaik.auth.structs.*
import at.tugraz.iaik.p2p.commons.CommonCrypto
import at.tugraz.iaik.p2p.commons.CryptoException
import iaik.asn1.structures.AlgorithmID
import iaik.security.ssl.CipherSuite
import iaik.security.ssl.SSLSocket
import java.io.IOException
import java.net.Socket
import java.security.PrivateKey
import java.security.cert.X509Certificate
import java.util.*

/**
 * Created by Bernd Prünster <bernd.pruenster@iaik.tugraz.at>
 */
open class TLSAuth @Throws(IOException::class, CryptoException::class) constructor(port: Int,
                                                                                   certChain: Array<X509Certificate>,
                                                                                   cipherSuites: Array<CipherSuite>,
                                                                                   validity: Validity,
                                                                                   crl: CRL,
                                                                                   signingKey: PrivateKey,
                                                                                   sigAlg: AlgorithmID) : AuthSocket(
        port,
        null,
        AuthResponder(signingKey, certChain[0], sigAlg, validity, crl),
        certChain,
        cipherSuites,
        signingKey) {
    init {
        server.context.requestClientCertificate = true
    }


    private class AuthResponder constructor(private val signingKey: PrivateKey,
                                            private val caCert: X509Certificate,
                                            private val sigAlg: AlgorithmID,
                                            private val validity: Validity,
                                            val crl: CRL) : AuthRequestListener {


        @Throws(CryptoException::class) override fun onRequest(src: Socket,
                                                               auth: AuthSocket,
                                                               request: Request): AuthResponse {

            if (request.type == RequestType.REVOKE) {
                val clientCert = (src as SSLSocket).peerCertificateChain[0]
                if (clientCert.subjectDN == request.principal) {
                    crl.revoke(clientCert, signingKey, sigAlg)
                    return AuthResponse(AuthResponse.Result.GRANTED, clientCert)
                }
                return AuthResponse(AuthResponse.Result.DENIED)
            }

            val to = Date()
            to.time += validity.unit.toMillis(validity.number.toLong())
            val cert = CommonCrypto.genCert(request.pubKey,
                                            request.principal,
                                            caCert.subjectDN,
                                            signingKey,
                                            sigAlg,
                                            false,
                                            Date(),
                                            to,
                                            crl.crlDistPoints)
            return AuthResponse(AuthResponse.Result.GRANTED, cert)
        }
    }
}