package at.tugraz.iaik.auth.mgmt.config

import at.tugraz.iaik.auth.io.GensonIO
import iaik.asn1.structures.AlgorithmID
import iaik.asn1.structures.Name
import iaik.security.ssl.CipherSuite
import java.io.File
import java.io.FileInputStream
import java.io.FileOutputStream
import java.io.FileReader
import java.lang.IllegalArgumentException
import java.security.Principal
import java.security.PrivateKey
import java.security.cert.X509Certificate
import java.util.*
import java.util.concurrent.TimeUnit


class AuthConfig(path: String = "auth.config") : Properties() {


    override fun getProperty(key: String, defaultValue: String): String {
        val value = super.getProperty(key, defaultValue);
        super.setProperty(key, value)
        return value
    }

    init {
        val r = FileReader(path)
        load(r)
        r.close()
    }


    fun getServiceDescriptor(name: String): ServiceDescriptor {
        val keystore = getProperty("${name.replace(".", "\\.")}.keystore", "$name.p12")
        val crlPoints = getProperty("${name.replace(".", "\\.")}.crlpoints", "[]")
        val points = GensonIO.deserialize(crlPoints.toByteArray(), Array<String>::class.java)
        return ServiceDescriptor(readName(name), readValidity(name), readSigAlg(name), keystore, points)
    }

    fun setServiceDescriptor(name: String, descriptor: ServiceDescriptor) {
        writeName(name, descriptor.name)
        writeValidity(name, descriptor.validity)
        writeSigAlg(name, descriptor.sigAlg)
        setProperty("${name.replace(".", "\\.")}.keystore", descriptor.keystore)
    }

    fun readAuthProps(name: String): MutableList<AuthProperties> {
        val authInfos = mutableListOf<AuthProperties>()
        for (k in keys) {
            val prefix = "${name.replace(".", "\\.")}.auth."
            if (k.toString().startsWith(prefix)) {
                val authProps = object : Properties() {
                    override fun getProperty(key: String, defaultValue: String): String {
                        val value = super.getProperty(key, defaultValue);
                        super.setProperty(key, value)
                        return value
                    }
                }
                val file = File(getProperty(k as String))

                authProps.load(FileInputStream(file))
                val type = AuthType.valueOf(authProps.getProperty("type", "TLS"))
                val port = authProps.getProperty("port", "9996").toInt()
                val read = authProps.getProperty("validity", "14 DAYS").split(" ")
                val validity = Validity(read[0].toInt(), TimeUnit.valueOf(read[1]))
                authInfos.add(AuthProperties(type, port, validity))
                val out = FileOutputStream(file, false)
                authProps.store(out, null)
                out.close()
            }
        }
        return authInfos
    }

    fun writeAuthProps(name: String, props: Map<String, AuthProperties>) {
        for (k in props) {
            setProperty("${name.replace(".", "\\.")}.auth.${k.key}", "${k.key}.conf")
            val p = Properties()
            p.setProperty("type", k.value.type.name)
            p.setProperty("port", k.value.port.toString())
            p.setProperty("validity", "${k.value.validity.number} ${k.value.validity.unit.name}")
            val out = FileOutputStream("${k.key}.conf")
            p.store(out, null)
            out.close()
        }
    }


    var keyStorePath: String
        get() = getProperty("keystore.path", "default.p12")
        set(value) {
            setProperty("keystore.path", value)
        }

    var keyStorePass: CharArray
        get() {
            val arr = getProperty("keystore.pass", "").toCharArray()
            if (arr.isEmpty()) throw RuntimeException("Please specify a password for unlocking the key store!")
            return arr
        }
        set(value) {
            setProperty("keystore.pass", String(value))
        }


    var caName: Principal
        get() = readName("ca")
        set(value) {
            writeName("ca", value)
        };

    private fun writeName(name: String, value: Principal) {
        setProperty("${name.replace(".", "\\.")}.name", value.toString())
    }

    private fun readName(name: String) = Name(getProperty("${name.replace(".", "\\.")}.name", "CN=$name"))

    var caSigAlg: SigAlg
        get() = readSigAlg("ca")
        set(value) {
            writeSigAlg("ca", value)
        }

    private fun readSigAlg(name: String) = SigAlg.getByID(getProperty("${name.replace(".", "\\.")}.sigalg",
                                                                      SigAlg.SHA256_RSA.algName))

    private fun writeSigAlg(name: String, alg: SigAlg) {
        setProperty("${name.replace(".", "\\.")}.sigalg", alg.algName)
    }

    var caValidity: Validity
        get() = readValidity("ca")
        set(value) {
            writeValidity("ca", value)
        }

    private fun readValidity(name: String): Validity {
        val read = getProperty("${name.replace(".", "\\.")}.validity", "365 DAYS").split(" ")
        return Validity(read[0].toInt(), TimeUnit.valueOf(read[1]))
    }

    private fun writeValidity(name: String, validity: Validity) {
        setProperty("${name.replace(".", "\\.")}.validity", "${validity.number.toString()} ${validity.unit.name}")
    }


    fun getSupportedCipherSuites(algorithm: String): Array<CipherSuite> {
        if (algorithm == "RSA") {
            return arrayOf(CipherSuite.getByName("TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384"),
                           CipherSuite.getByName("TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256"),
                           CipherSuite.getByName("TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256"))

        } else if (algorithm == "EC") {
            return arrayOf(CipherSuite.getByName("TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA384"),
                           CipherSuite.getByName("TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256"),
                           CipherSuite.getByName("TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA256"))
        }
        throw IllegalArgumentException("Algorithm $algorithm is not supported")
    }
}

enum class SigAlg(val algName: String, val algorithm: AlgorithmID, val keyType: String = "EC", val len: Int = 256) {
    SHA3_ECDSA("ECDSA/SHA3", AlgorithmID.ecdsa_With_SHA3_256),
    SHA256_ECDSA("ECDSA/SHA256", AlgorithmID.ecdsa_With_SHA256),
    SHA256_RSA("RSA/SHA256", AlgorithmID.sha256WithRSAEncryption, "RSA", 2048);

    companion object {
        @JvmStatic
        fun getByID(name: String): SigAlg {
            for (alg in SigAlg.values()) {
                if (alg.algName == name) return alg
            }
            throw RuntimeException("Algorithm $name is not defined!")
        }
    }
}

enum class AuthType {
    LOCAL,
    QR,
    TLS,
    QR_TLS
}

data class CertAndKey(val cert: X509Certificate, val key: PrivateKey)

data class Validity(val number: Int, val unit: TimeUnit)


data class ServiceDescriptor(val name: Principal,
                             val validity: Validity,
                             val sigAlg: SigAlg,
                             val keystore: String,
                             val crlDistributionPoints: Array<String>)

data class AuthProperties(val type: AuthType, val port: Int, val validity: Validity)