package at.tugraz.iaik.auth.ocsp

import iaik.asn1.structures.AlgorithmID
import iaik.asn1.structures.Name
import iaik.security.ssl.OCSPStatusRequest
import iaik.x509.V3Extension
import iaik.x509.extensions.ReasonCode
import iaik.x509.ocsp.*
import iaik.x509.ocsp.utils.ResponseGenerator
import java.io.ByteArrayInputStream
import java.io.ByteArrayOutputStream
import java.security.PrivateKey
import java.security.cert.X509Certificate
import java.util.*


class OCSPResponder(private val signingKey: PrivateKey, private val chain: Array<X509Certificate>) {

    val respGen = ResponseGenerator(signingKey, convertChain(chain))

    init {
        respGen.printDebug(true)
    }

    fun respond(ocspStatusRequest: OCSPStatusRequest, cert: X509Certificate): OCSPResponse {
        val certID = CertID(AlgorithmID.sha3_256,
                            chain[chain.size - 1].subjectDN as Name,
                            chain[chain.size - 1].publicKey,
                            cert.serialNumber)

        val reqCert = ReqCert(ReqCert.certID, certID)
        val request = Request(reqCert)
        val ocspRequest = OCSPRequest()
        ocspRequest.requestList = arrayOf(request)
        for (ext in ocspStatusRequest.extensions.listExtensions()) {
            ocspRequest.addExtension(ext as V3Extension)
        }

        val date = Date()
        val reasonCode = ReasonCode(ReasonCode.keyCompromise)
        val ri = RevokedInfo(date)
        ri.revocationReason = reasonCode

        val s = SingleResponse(reqCert, CertStatus(), Date())
        respGen.addResponseEntry(s)

        respGen.addCertificateIssuer(chain[chain.size - 1].subjectDN as Name)

        return respGen.createOCSPResponse(ByteArrayInputStream(ocspRequest.encoded),
                                          null,
                                          AlgorithmID.ecdsa_With_SHA3_256,
                                          null)

    }

    private fun convertChain(chain: Array<X509Certificate>): Array<iaik.x509.X509Certificate> {
        return Array(chain.size, { chain[it].convert() })
    }

}


fun X509Certificate.convert(): iaik.x509.X509Certificate {
    return iaik.x509.X509Certificate(encoded)
}


fun OCSPResponse.encode(): ByteArray {
    val out = ByteArrayOutputStream(encoded.size + 3)
    out.write(encoded.size.shr(16).toByte().toInt())
    out.write(encoded.size.shr(8).toByte().toInt())
    out.write(encoded.size.toByte().toInt())
    out.write(encoded)
    return out.toByteArray()
}