package at.tugraz.iaik.auth.socket

import at.tugraz.iaik.auth.ocsp.OCSPResponder
import iaik.security.ssl.*
import java.net.InetAddress
import java.security.PrivateKey
import java.security.cert.X509Certificate


/**
 * Created by Bernd Prünster <bernd.pruenster@iaik.tugraz.at>
 */


private fun createDefaultExtensions(): ExtensionList {

    val extensions = ExtensionList()

    // Supported Elliptic Curves
    val supportedEllipticCurves = SupportedEllipticCurves()
    extensions.addExtension(supportedEllipticCurves)

    // Supported Point Formats
    val supportedPointFormats = SupportedPointFormats(arrayOf(SupportedPointFormats.PF_COMPRESSED_CHAR2,
                                                              SupportedPointFormats.PF_COMPRESSED_PRIME,
                                                              SupportedPointFormats.PF_UNCOMPRESSED))
    extensions.addExtension(supportedPointFormats)


    extensions.setAllCritical(false)

    // enable exensions
    return extensions

}

class TLSServerSocket(host: InetAddress? = null,
                      port: Int,
                      chain: Array<X509Certificate>,
                      signingKey: PrivateKey,
                      cipherSuites: Array<CipherSuite>,
                      responder: OCSPResponder? = null) : SSLServerSocket(port,
                                                                          50,
                                                                          host,
                                                                          setupSrvCtx(chain,
                                                                                      signingKey,
                                                                                      cipherSuites,
                                                                                      responder)) {

    companion object {
        init {
            SecurityProvider.setSecurityProvider(ECCelerateProvider());
        }

        private fun setupSrvCtx(chain: Array<X509Certificate>,
                                signingKey: PrivateKey,
                                cipherSuites: Array<CipherSuite>,
                                responder: OCSPResponder? = null): SSLServerContext {
            val ctx = SSLServerContext()
            //  ctx.chainVerifier = OCSPCertStatusChainVerifier()
            ctx.addServerCredentials(object : KeyAndCert(chain, signingKey) {
//                override fun getCertificateStatus(p0: Int, p1: ByteArray?, p2: SSLTransport?): ByteArray? {
//                    System.err.println("SRV RESP")
//                    return responder?.respond(OCSPStatusRequest(p1), chain[0])?.encode()
//                }
            })
            ctx.setAllowedProtocolVersions(SSLContext.VERSION_TLS12, SSLContext.VERSION_TLS12)
            ctx.enabledCipherSuites = cipherSuites
            ctx.updateCipherSuites()
            ctx.allowLegacyRenegotiation = false
            ctx.addTrustedCertificate(chain[chain.size - 1])
            val ext = createDefaultExtensions()
            //   val ocsp = CertificateStatusRequest(OCSPStatusRequest.STATUS_TYPE, null)
            // ocsp.setCritical(true)
            // ext.addExtension(ocsp)
            ctx.setExtensions(ext)
            return ctx
        }
    }

}


class TLSSocket(host: InetAddress,
                port: Int,
                chain: Array<X509Certificate>? = null,
                clientKey: PrivateKey? = null,
                trustedCerts: Collection<X509Certificate>? = null,
                cipherSuites: Array<CipherSuite>,
                responder: OCSPResponder? = null) : SSLSocket(host,
                                                              port,
                                                              setupClCtx(chain,
                                                                         clientKey,
                                                                         trustedCerts,
                                                                         cipherSuites,
                                                                         responder)) {

    companion object {
        init {
            SecurityProvider.setSecurityProvider(ECCelerateProvider())
        }

        private fun setupClCtx(chain: Array<X509Certificate>?,
                               clientKey: PrivateKey?,
                               trustedCerts: Collection<X509Certificate>?,
                               cipherSuites: Array<CipherSuite>,
                               responder: OCSPResponder? = null): SSLContext? {
            if (chain != null && clientKey == null) {
                throw IllegalArgumentException("Key is required whenever a certificate is provided!")
            }
            val ctx = SSLClientContext()
            //  ctx.chainVerifier = OCSPCertStatusChainVerifier()
            trustedCerts?.forEach { ctx.addTrustedCertificate(it) }
            chain?.let {
                ctx.addClientCredentials(chain, clientKey)
            }

            chain?.let {
                ctx.addClientCredentials(object : KeyAndCert(chain, clientKey) {
//                    override fun getCertificateStatus(p0: Int, p1: ByteArray?, p2: SSLTransport?): ByteArray? {
//                        System.err.println("CLI RESP")
//                        return responder?.respond(OCSPStatusRequest(p1), chain[0])?.encode()
//                    }
                })
            }
            ctx.setAllowedProtocolVersions(SSLContext.VERSION_TLS12, iaik.security.ssl.SSLContext.VERSION_TLS12)
            ctx.enabledCipherSuites = cipherSuites
            ctx.updateCipherSuites()
            ctx.allowLegacyRenegotiation = false
            val ext = createDefaultExtensions()
            //  val ocsp = CertificateStatusRequest(OCSPStatusRequest.STATUS_TYPE, null)
            // ocsp.setCritical(true)
            // ext.addExtension(ocsp)
            ctx.setExtensions(ext)
            return ctx
        }
    }
}


