import at.tugraz.iaik.auth.impl.LocalAuth
import at.tugraz.iaik.auth.impl.QRAuth
import at.tugraz.iaik.auth.impl.TLSAuth
import at.tugraz.iaik.auth.io.GensonIO
import at.tugraz.iaik.auth.mgmt.config.*
import at.tugraz.iaik.auth.ocsp.convert
import at.tugraz.iaik.auth.socket.AuthSocket
import at.tugraz.iaik.auth.socket.TLSSocket
import at.tugraz.iaik.auth.structs.AuthResponse
import at.tugraz.iaik.auth.structs.CRL
import at.tugraz.iaik.auth.structs.Request
import at.tugraz.iaik.auth.structs.RequestType
import at.tugraz.iaik.p2p.commons.CommonCrypto
import iaik.asn1.ObjectID
import iaik.asn1.structures.Name
import java.io.File
import java.io.FileInputStream
import java.io.FileOutputStream
import java.io.FileWriter
import java.net.InetAddress
import java.security.KeyStore
import java.security.Principal
import java.security.PrivateKey
import java.security.cert.X509Certificate
import java.util.*
import java.util.concurrent.TimeUnit

fun setupCA(caName: Principal, validity: Validity, algorithm: SigAlg): CertAndKey {
    val keyPair = CommonCrypto.genKP(algorithm.keyType, algorithm.len)
    var to = Date();
    to.time = to.time + validity.unit.toMillis(validity.number.toLong())
    val caCert = CommonCrypto.genCert(pub = keyPair.public,
                                      subject = caName,
                                      signingKey = keyPair.private,
                                      sigAlg = algorithm.algorithm,
                                      ca = true,
                                      from = Date(),
                                      to = to)

    return CertAndKey(caCert, keyPair.private)
}

fun setupService(name: Principal, ca: CertAndKey, algorithm: SigAlg, validity: Validity): Array<CertAndKey> {
    val intermedKeyPair = CommonCrypto.genKP(algorithm.keyType, algorithm.len)
    val keyPair = CommonCrypto.genKP(algorithm.keyType, algorithm.len)
    var to = Date();

    val intermedName = Name(name.toString())
    val rdn = intermedName.getRDN(ObjectID.commonName)
    intermedName.removeRDN(ObjectID.commonName)
    intermedName.addRDN(ObjectID.commonName, "$rdn intermediate CA")

    to.time = to.time + validity.unit.toMillis(validity.number.toLong())

    val caCert = CommonCrypto.genCert(intermedKeyPair.public,
                                      intermedName,
                                      ca.cert.subjectDN,
                                      ca.key,
                                      ca.cert.convert().signatureAlgorithm,
                                      true,
                                      Date(),
                                      to)
    val cert = CommonCrypto.genCert(keyPair.public,
                                    name,
                                    caCert.subjectDN,
                                    intermedKeyPair.private,
                                    caCert.convert().signatureAlgorithm,
                                    false,
                                    Date(),
                                    to)

    return arrayOf(CertAndKey(cert, keyPair.private), CertAndKey(caCert, intermedKeyPair.private))
}


fun setup(path: String, passwd: CharArray, keyStorePath: String) {
    CommonCrypto.init()
    val file = File(path)
    file.createNewFile()
    println("Using config file ${file.absolutePath}")

    val config = AuthConfig(path)

    config.keyStorePath = keyStorePath
    config.keyStorePass = passwd
    val keyStore = KeyStore.getInstance("PKCS12", "IAIK");
    keyStore.load(null, null)

    val ca = setupCA(config.caName, config.caValidity, config.caSigAlg)
    val vpnConfig = config.getServiceDescriptor("DemoVPN")
    val vpn = setupService(vpnConfig.name, ca, vpnConfig.sigAlg, vpnConfig.validity)

    keyStore.setKeyEntry(ca.cert.subjectDN.name, ca.key, passwd, arrayOf(ca.cert))
    keyStore.setKeyEntry(vpn[1].cert.subjectDN.name, vpn[1].key, passwd, arrayOf(vpn[1].cert, ca.cert))

    val vpnKS = KeyStore.getInstance("PKCS12", "IAIK")
    vpnKS.load(null, null)

    vpnKS.setKeyEntry(vpn[0].cert.subjectDN.name, vpn[0].key, passwd, arrayOf(vpn[0].cert, vpn[1].cert, ca.cert))

    val vpnOut = FileOutputStream(vpnConfig.keystore)
    vpnKS.store(vpnOut, passwd)
    vpnOut.close()

    config.setProperty("DemoVPN.crlpoints", "[\"http://localhost:8080/demovpn.crl\"]")

    config.writeAuthProps("DemoVPN",
                          mapOf(Pair("localauth", AuthProperties(AuthType.LOCAL, 9995, Validity(14, TimeUnit.DAYS))),
                                Pair("tlsauth", AuthProperties(AuthType.TLS, 9996, Validity(14, TimeUnit.DAYS)))))

    val keyOutStream = FileOutputStream(keyStorePath)
    keyStore.store(keyOutStream, passwd)
    keyOutStream.close()
    val confWriter = FileWriter(file, false)
    config.store(confWriter, null)

    confWriter.close()
}

fun authDemo(path: String) {

    //load config
    CommonCrypto.init()
    val config = AuthConfig(path)
    val name = "DemoVPN"
    val vpn = config.getServiceDescriptor(name)
    val keyStore = KeyStore.getInstance("PKCS12", "IAIK")
    keyStore.load(FileInputStream(config.keyStorePath), config.keyStorePass)
    val auth = config.readAuthProps(name)
    val intermedName = Name(vpn.name.toString())
    val rdn = intermedName.getRDN(ObjectID.commonName)
    intermedName.removeRDN(ObjectID.commonName)
    intermedName.addRDN(ObjectID.commonName, "$rdn intermediate CA")

    val chain = keyStore.getCertificateChain(intermedName.name)
    val authSocket = Array<AuthSocket>(auth.size, {
        val key = keyStore.getKey(intermedName.name, config.keyStorePass)
        val cipherSuites = config.getSupportedCipherSuites(key.algorithm)
        val sock: AuthSocket
        when (auth[it].type) {
            AuthType.LOCAL  -> {
                sock = LocalAuth(auth[it].port,
                                 chain as Array<X509Certificate>,
                                 cipherSuites,
                                 vpn.validity,
                                 vpn.crlDistributionPoints,
                                 key as PrivateKey,
                                 vpn.sigAlg.algorithm)
            }
            AuthType.QR     -> {
                sock = QRAuth(auth[it].port,
                              chain as Array<X509Certificate>,
                              cipherSuites,
                              vpn.validity,
                              vpn.crlDistributionPoints,
                              key as PrivateKey,
                              vpn.sigAlg.algorithm)
            }
            AuthType.QR_TLS -> {
                sock = QRAuth(auth[it].port,
                              chain as Array<X509Certificate>,
                              cipherSuites,
                              vpn.validity,
                              vpn.crlDistributionPoints,
                              key as PrivateKey,
                              vpn.sigAlg.algorithm,
                              true)
            }
            AuthType.TLS    -> {
                sock = TLSAuth(auth[it].port,
                               chain as Array<X509Certificate>,
                               cipherSuites,
                               vpn.validity,
                               CRL(chain[0].issuerDN, File("$name.crl"), vpn.crlDistributionPoints),
                               key as PrivateKey,
                               vpn.sigAlg.algorithm)
            }
        }
        Thread(sock).start()
        sock
    })


    //prepare for cert request
    val clientKP = CommonCrypto.genKP("RSA", 2048)

    //listen
    val socket = TLSSocket(InetAddress.getLoopbackAddress(),
                           9995,
                           trustedCerts = listOf(chain[0] as X509Certificate),
                           responder = authSocket[0].responder,
                           cipherSuites = config.getSupportedCipherSuites("RSA"))

    val outS = (socket.outputStream)

    //send request
    GensonIO.serialize(Request(clientKP.public, Name("CN=dummy"), RequestType.AUTH), outS)
    outS.flush()


    val resp = GensonIO.deserialize(socket.inputStream, AuthResponse::class.java)

    //print result
    System.err.println(resp)

    //revoke
    if (resp.state == AuthResponse.Result.GRANTED) {
        val certChain = arrayOf(resp.cert!!) + (chain as Array<X509Certificate>)
        val socket = TLSSocket(InetAddress.getLoopbackAddress(),
                               9996,
                               trustedCerts = listOf(chain[0] as X509Certificate),
                               chain = certChain,
                               clientKey = clientKP.private,
                               responder = authSocket[1].responder,
                               cipherSuites = config.getSupportedCipherSuites("RSA"))
        val outS = (socket.outputStream)

        //send revocation request
        GensonIO.serialize(Request(clientKP.public, Name("CN=dummy"), RequestType.REVOKE), outS)
        outS.flush()
        val resp = GensonIO.deserialize(socket.inputStream, AuthResponse::class.java)

        //print result
        System.err.println(resp)
    }


}


fun main(args: Array<String>) {
    //setup
    setup("auth.config", charArrayOf('s', 'e', 'c', 'r', 'e', 't'), "demo.p12")
    //demo
    authDemo("auth.config")
    System.exit(0)
}
